# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from functools import wraps

import mock


def patch_class_method_with_original(cls, method_name):
    """
    Mock a class method without changing the method return value.
    Useful for patching methods where method arguments have to be
    checked however return value should not be mocked.

    For example:

    class Foo(object):
        def a(self):
            return self.b(123) * 2
        def b(self, arg):
            return arg

    class TestFoo(TestCase):
        @patch_class_method_with_original(Foo, 'b')
        def test_a(self, mock_b):
            foo = Foo()
            actual = foo.a()
            self.assertEqual(actual, 246)
            mock_b.assert_called_once_with(123)

    """
    original_mock = getattr(cls, method_name)

    def outer_wrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            with mock.patch.object(cls, method_name, autospec=True) as mocked:
                mocked.side_effect = original_mock
                new_args = args + (mocked,)
                return f(*new_args, **kwargs)

        return wrapper

    return outer_wrapper



from __future__ import print_function, unicode_literals
import inspect

from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import Resolver404, resolve
from django.template.response import SimpleTemplateResponse
from six.moves.urllib_parse import urlsplit, urlunsplit


class BaseValidator(object):
    """
    Base validator which should be sub-classed to create
    custom validators.

    :var tuple BaseValidator.expected_attrs: Required attributes for the
        validator. :py:meth:`_check_improper_configuration`
        will verify that all of these attributes are defined.

        .. note:: Each validator should only define required
            attributes for itself. :py:meth:`_get_expected_attrs`
            will automatically return required attributes
            from all current validator and base classes.

    :param test_step: :py:class:`subui.step.TestStep`
        instance which will be used to make assertions on.

        .. note:: This parameter is not really required in the
            ``__init__`` because the same step will be passed
            in :py:meth:`test` however is useful in ``__init__``
            in case subclass validator needs to apply custom
            login according to values from the ``step``.
    """
    expected_attrs = None

    def __init__(self, test_step=None, **kwargs):
        self.step = test_step
        self.__dict__.update(**kwargs)

    def _get_expected_attrs(self):
        """
        Get all required/expected attributes as defined by
        :py:attr:`expected_attrs` from all current and base
        classes.

        For example::

            class Validator1(BaseValidator):
                expected_attr = ('foo', 'bar')

            class Validator2(Validator1):
                expected_attr = ('hello', 'world')

            > validator = Validator2()
            > validator._get_expected_attrs()
            ['foo', 'bar', 'hello', 'world']

        :rtype: list
        """
        bases = inspect.getmro(self.__class__)
        attrs = set()
        for base in bases:
            attrs |= set(getattr(base, 'expected_attrs', None) or set())
        return list(sorted(attrs))

    def _check_improper_configuration(self):
        """
        Check that the validator is configured correctly
        by verifying that all attributes as returned by
        :py:meth:`_get_expected_attrs` are defined.

        :raises ImproperlyConfigured: If any of the required
            attributes are not defined.
        """
        for attr in self._get_expected_attrs():
            if not getattr(self, attr, None):
                msg = '{} requires to define {}'
                raise ImproperlyConfigured(
                    msg.format(self.__class__.__name__, attr)
                )

    def _get_base_error_message(self):
        """
        Get base error message which will be used in assertions
        """
        msg = 'Response for {{{key}:{step}}} requesting "{url}"'
        msg = msg.format(
            key=self.step.step_key,
            step=self.step.__class__.__name__,
            url=self.step.get_url(),
        )
        return msg

    def test(self, test_step):
        """
        Test the step's server response by making
        all the necessary assertions. This method
        by default saves the ``test_step`` parameter
        into :py:attr:`step` and validates the validator
        by using :py:meth:`_check_improper_configuration`.
        All subclass validators should actually implement
        assertions in this method.

        :param test_step: Test step
        """
        self.step = test_step
        self._check_improper_configuration()


class HeaderValidator(BaseValidator):
    """
    Validator which can check that a particular header
    is returned and that it is of particular value.

    :var str HeaderValidator.header_name: Name of the header which
        must be returned
    :var str expected_header: Expected header value
        to be returned
    :var bool HeaderValidator.test_header_value: Whether to test the
        header value or simply check its existence
    :var test_contains_value: Whether to test if the
        header value contains another value, or simply
        check equality to that value
    """
    expected_attrs = ('header_name', 'expected_header',)
    header_name = None
    expected_header = None
    test_header_value = True
    test_contains_value = False

    def _get_expected_attrs(self):
        """
        Get all expected attributes except remove "expected_header"
        if :py:attr:`test_header_value` is ``False``.
        """
        attrs = super(HeaderValidator, self)._get_expected_attrs()

        if not self.test_header_value:
            attrs.remove('expected_header')

        return attrs

    def test(self, test_step):
        """
        Test the response returned with
        :py:attr:header_name header and that its value
        is equal to :py:attr:expected_header if
        :py:attr:test_header_value is True.
        If :py:attr:test_contains_value is True,
        header value will be tested to contain expected value.
        """
        super(HeaderValidator, self).test(test_step)
        if self.test_contains_value:
            self.test_header_value = False
        self.step.test.assertIn(
            self.header_name,
            self.step.response,
            '{} must contain {} header'
            ''.format(self._get_base_error_message(),
                      self.header_name)
        )
        if self.test_header_value:
            self.test_contains_value = False
            self.step.test.assertEqual(
                self.step.response[self.header_name],
                self.expected_header,
                '{} returned header {} with value {} != {}'
                ''.format(self._get_base_error_message(),
                          self.header_name,
                          self.step.response[self.header_name],
                          self.expected_header)
            )
        if self.test_contains_value:
            self.step.test.assertIn(
                self.expected_header,
                self.step.response[self.header_name],
                '{} returned header {} with value {} which doesnt contain {}'
                ''.format(self._get_base_error_message(),
                          self.header_name,
                          self.step.response[self.header_name],
                          self.expected_header)
            )


class HeaderContentTypeValidator(HeaderValidator):
    """
    Validator to check that the expected
    "Content-Type" header is returned.
    """
    header_name = 'Content-Type'


class HeaderLocationValidator(HeaderValidator):
    """
    Validator to check that the redirect
    "Location" header is returned
    """
    header_name = 'Location'


class StatusCodeValidator(BaseValidator):
    """
    Validator which allows to verify the returned
    server status code such as "OK-200" or "Redirect-302", etc.

    :var int StatusCodeValidator.expected_status_code: Expected status code to be
        returned by the server
    """
    expected_attrs = ('expected_status_code',)
    expected_status_code = None

    def test(self, test_step):
        """
        Test that the response status code
        matched expected status code.
        """
        super(StatusCodeValidator, self).test(test_step)

        self.step.test.assertEqual(
            self.step.response.status_code,
            self.expected_status_code,
            '{} returned with status code {} != {}'
            ''.format(self._get_base_error_message(),
                      self.step.response.status_code,
                      self.expected_status_code)
        )


class StatusCodeOkValidator(StatusCodeValidator):
    """
    Validator to check that the returned status
    code is OK - 200
    """
    expected_status_code = 200


class StatusCodeRedirectValidator(HeaderLocationValidator, StatusCodeValidator):
    """
    Validator to check that the server returns a redirect
    with the "Location" header defined.
    """
    expected_status_code = 302
    test_header_value = False


class RedirectToRouteValidator(StatusCodeRedirectValidator):
    """
    Validator which also checks that the server returns
    a redirect to an expected Django route.

    :var str expected_route_name: Route name to which
        the server should redirect to
    """
    expected_attrs = ('expected_route_name',)
    expected_route_name = None

    def test(self, test_step):
        """
        Test the response by additionally testing
        that the response redirects to an expected
        route as defined by :py:attr:`expected_route_name`.
        """
        super(RedirectToRouteValidator, self).test(test_step)

        location = self.step.response['Location']
        # remove schema, query string and host from URL. Since query string need to be removed to properly resolve that url
        location = urlunsplit(('', '', urlsplit(location).path, '', ''))

        try:
            redirected_to_route = resolve(location).view_name
        except Resolver404:
            msg = '{} returned a redirect to "{}" which cannot be resolved'
            self.step.test.fail(msg.format(self._get_base_error_message(),
                                           location))

        self.step.test.assertEqual(
            redirected_to_route,
            self.expected_route_name,
            '{} returned redirect to route {} != {}'
            ''.format(self._get_base_error_message(),
                      redirected_to_route,
                      self.expected_route_name)
        )


class ResponseContentContainsValidator(StatusCodeOkValidator):
    """
    Validator which also checks that returned response
    content contains expect string.

    :var str expected_content: Expected string in the
        server response
    """
    expected_attrs = ('expected_content',)
    expected_content = None

    def test(self, test_step):
        """
        Test the response by additionally testing
        that the response context contains expected
        string as defined by :py:attr:`expected_content`.
        """
        super(ResponseContentContainsValidator, self).test(test_step)

        self.step.test.assertIn(
            self.expected_content,
            # need to decode since content is binary
            self.step.response.content.decode('utf-8'),
            '{} does not contain {!r} in its content'
            ''.format(self._get_base_error_message(),
                      self.expected_content)
        )


class ResponseContentNotContainsValidator(StatusCodeOkValidator):
    """
    Validator checks that returned response
    content does not contain unexpected string.

    :var str unexpected_content: Unexpected string in the
        server response
    """
    expected_attrs = ('unexpected_content',)
    unexpected_content = None

    def test(self, test_step):
        """
        Test the response by additionally testing
        that the response context does not contain the unexpected
        string as defined by :py:attr:`unexpected_content`.
        """
        super(ResponseContentNotContainsValidator, self).test(test_step)

        self.step.test.assertNotIn(
            self.unexpected_content,
            # need to decode since content is binary
            self.step.response.content.decode('utf-8'),
            'UnexpectedContentFound: {} contains {!r} in its content. '
            ''.format(self._get_base_error_message(),
                      self.unexpected_content)
        )


class SessionDataValidator(BaseValidator):
    """
    Validator which allows to verify the data in the session based on
    session key.

    :var str expected_session_key: Expected session key to be present in session
    :var list expected_session_secondary_keys: List of Expected session key to be present in session
    """
    expected_attrs = ('expected_session_key',)
    expected_session_key = None
    expected_session_secondary_keys = []

    def test(self, test_step):
        """
        Test that expected session key is present. If expected session data provided,
        ensure the expected session key data matches what is there currently.
        """
        super(SessionDataValidator, self).test(test_step)
        self.step.test.assertIn(
            self.expected_session_key,
            self.step.response.wsgi_request.session.keys(),
            '{} does not contain session[{!r}].'
            ''.format(self._get_base_error_message(),
                      self.expected_session_key)
        )
        if self.expected_session_secondary_keys:
            self.step.test.assertIsInstance(
                self.step.response.wsgi_request.session[self.expected_session_key],
                dict,
                '{} session[{!r}] is not a dictionary hence cannot contain secondary keys.'
                ''.format(self._get_base_error_message(), self.expected_session_key)
            )
        # Make sure secondary keys are not empty.
        for secondary_key in self.expected_session_secondary_keys:
            self.step.test.assertIn(
                secondary_key,
                self.step.response.wsgi_request.session[self.expected_session_key].keys(),
                '{} does not contain session[{!r}][{!r}].'
                ''.format(self._get_base_error_message(), self.expected_session_key, secondary_key)
            )
            self.step.test.assertIsNotNone(
                self.step.response.wsgi_request.session[self.expected_session_key][secondary_key],
                '{} contains session[{!r}][{!r}] but is empty.'
                ''.format(self._get_base_error_message(), self.expected_session_key, secondary_key)
            )


class FormInitialDataValidator(BaseValidator):
    """
    Validator checks that form in response has expected data in initial data.

    :var str initial_data_key: Expected initial data key to be present in form initial data
    :var expected_initial_data_value: Expected value initial value should be set to
    :var str context_data_form_name: Template context data key for form data
    :var bool test_initial_value: Test if the initial value matched expected value
    :var bool test_initial_value_present: Test if the initial value key is present in initial data
    :var bool test_initial_value_not_none: Test if the initial value is not ``None``
    """
    expected_attrs = ('initial_data_key',)
    initial_data_key = None
    expected_initial_data_value = None
    context_data_form_name = 'form'

    test_initial_value = False
    test_initial_value_present = True
    test_initial_value_not_none = True

    def test(self, test_step):
        super(FormInitialDataValidator, self).test(test_step)

        self.step.test.assertIsInstance(
            self.step.response,
            SimpleTemplateResponse,
            '{} did not return SimpleTemplateResponse '
            'and as such response.context_data is not accessible. '
            'It returned {!r}'
            ''.format(self._get_base_error_message(),
                      type(self.step.response))
        )

        self.step.test.assertIn(
            self.context_data_form_name,
            self.step.response.context_data,
            '{} did not render with {!r} in its context_data'
            ''.format(self._get_base_error_message(),
                      self.context_data_form_name)
        )

        self.step.test.assertIsInstance(
            self.step.response.context_data[self.context_data_form_name].initial,
            dict,
            '{} did not render with the context_data[{!r}].initial being a dictionary'
            ''.format(self._get_base_error_message(),
                      self.context_data_form_name)
        )

        if self.test_initial_value_present:
            self.step.test.assertIn(
                self.initial_data_key,
                self.step.response.context_data[self.context_data_form_name].initial.keys(),
                '{} did not render with {!r} in the context_data[{!r}].initial. '
                'Provided keys - {!r}'
                ''.format(self._get_base_error_message(),
                          self.initial_data_key,
                          self.context_data_form_name,
                          self.step.response.context_data[self.context_data_form_name].initial.keys())
            )

        if self.test_initial_value_not_none:
            self.step.test.assertIsNotNone(
                self.step.response.context_data['form'].initial[self.initial_data_key],
                '{} rendered with {!r} for context_data[{!r}].initial[{!r}]'
                ''.format(self._get_base_error_message(),
                          None,
                          self.context_data_form_name,
                          self.initial_data_key)
            )

        if self.test_initial_value:
            self.step.test.assertEqual(
                self.step.response.context_data['form'].initial[self.initial_data_key],
                self.expected_initial_data_value,
                '{} rendered with context_data[{!r}].initial[{!r}] = {} != {}'
                ''.format(self._get_base_error_message(),
                          self.context_data_form_name,
                          self.initial_data_key,
                          self.step.response.context_data['form'].initial[self.initial_data_key],
                          self.expected_initial_data_value)
            )



from __future__ import print_function, unicode_literals
import inspect
from collections import OrderedDict

from pycontext.context import Context


class SubUITestRunner(object):
    """
    SubUI Test Runner.

    This is the interface class of the SubUI framework.
    It runs all of the provided steps in the provided order.
    Since some steps might need state from previous executed
    steps, the runner maintains reference to a global state
    which it then passes to each step during execution.

    :var collections.OrderedDict steps: Test steps to be executed
    :var client: Django test client to run tests
    :var dict kwargs: Additional kwargs given to class.
        These kwargs will be provided to each step when executed.
    :var pycontext.context.Context state: State to be
        shared between test step executions.

    :param steps: Steps to be executed. Executed in the provided order
        hence need to be provided in order-maintaining data-structure.
    :type steps: list, tuple, collections.OrderedDict

    :param client: Django test Client to query server with
    :type client: django.test.client.Client

    :param state: State to be shared between step executions.
        Optional and by default is empty dict.
    :type state: dict

    :param kwargs: Additional kwargs to be passed to each
        step during initialization if it is not already provided
        as initialized object.
    """

    def __init__(self, steps, client, state=None, **kwargs):
        msg = 'SubUI steps can either be tuple, list or OrderedDict'
        assert isinstance(steps, (list, tuple, OrderedDict)), msg

        self.steps = steps
        self.client = client
        self.kwargs = kwargs

        self.state = Context(state or {})

        self._normalize_steps()

    def _normalize_steps(self):
        """
        Normalize steps - convert to :py:class:`collections.OrderedDict`
        and instantiate.

        This converts steps into :py:class:`collections.OrderedDict` if they are
        not already in which case the keys will be integer indexes of each step.
        Also this instantiates each step if is provided as a class.
        While instantiating it provides all the
        :py:attr:`SubUITestRunner.kwargs` given to runner.
        """
        if isinstance(self.steps, (list, tuple)):
            self.steps = OrderedDict(zip(range(len(self.steps)), self.steps))

        self.steps = OrderedDict(map(
            lambda s: (s[0], s[1](**self.kwargs)) if inspect.isclass(s[1]) else s,
            self.steps.items()
        ))

    def run(self):
        """
        Run the test runner.

        This executes all steps in the order there are defined
        in :py:attr:`SubUITestRunner.steps`. Before each step is executed
        :py:meth:`subui.step.TestStep.init`
        is called providing all the necessary attributes to execute
        the step like test client, steps, and state).

        :return: Reference to the test runner
        """
        for i, (key, step) in enumerate(self.steps.items()):
            step.init(client=self.client,
                      steps=self.steps,
                      step_index=i,
                      step_key=key,
                      state=self.state)
            step.request()
            step.test_response()

        return self



from __future__ import print_function, unicode_literals
from collections import OrderedDict
from unittest import TestCase

import mock
from pycontext.context import Context

from subui.test_runner import SubUITestRunner


TESTING_MODULE = 'subui.test_runner'


class TestSubUITestRunner(TestCase):
    @mock.patch.object(SubUITestRunner, '_normalize_steps')
    def test_init(self, mock_normalize_steps):
        """
        Test that init correctly handles input parameters
        """
        runner = SubUITestRunner(steps=[],
                                 client=mock.sentinel.client,
                                 state={'hello': 'world'})

        self.assertEqual(runner.steps, [])
        self.assertEqual(runner.client, mock.sentinel.client)
        self.assertIsInstance(runner.state, Context)
        self.assertEqual(runner.state, {'hello': 'world'})
        mock_normalize_steps.assert_called_once_with()

        with self.assertRaises(AssertionError):
            SubUITestRunner(steps=None, client=None)

    @mock.patch('inspect.isclass')
    def test_normalize_keys(self, mock_isclass):
        """
        Test that _normalize_steps converts steps to OrderedDict
        and instantiated each step if necessary
        """
        mock_isclass.side_effect = True, False
        step1 = mock.MagicMock()
        step2 = mock.MagicMock()

        with mock.patch.object(SubUITestRunner, '_normalize_steps'):
            runner = SubUITestRunner([step1, step2],
                                     None,
                                     something='here',
                                     morestuff='here')

        self.assertNotIsInstance(runner.steps, OrderedDict)

        runner._normalize_steps()

        step1.assert_called_once_with(something='here', morestuff='here')
        self.assertFalse(step2.called)

        self.assertIsInstance(runner.steps, OrderedDict)
        self.assertEqual(runner.steps, OrderedDict((
            (0, step1()),
            (1, step2),
        )))

    def test_run(self):
        """
        Test that run loops all steps and executes them
        """
        step = mock.MagicMock()
        runner = SubUITestRunner([step], None)

        actual = runner.run()

        self.assertIs(actual, runner)
        step.init.assert_called_once_with(
            client=runner.client,
            steps=runner.steps,
            step_index=0,
            step_key=0,
            state=runner.state
        )
        step.request.assert_called_once_with()
        step.test_response.assert_called_once_with()



from __future__ import print_function, unicode_literals
import inspect
import sys
import unittest
from collections import OrderedDict

import six
from django.core.urlresolvers import reverse

from .validators import BaseValidator


class TestStep(object):
    """
    Test step for :py:class:`subui.test_runner.SubUITestRunner`.

    The step is responsible for executing a self-contained task
    such as submitting a form to a particular URL and then
    make assertions regarding the server response.

    :var unittest.TestCase TestStep.test: Test class instance
        with which validators are going to run all their
        assertions with. By default it is an instance of
        :py:class:`unittest.TestCase` however can be changed
        to any other class to add additional assertion methods.
    :var str url_name: Name of the url as defined in ``urls.py``
        by which the URL is going to be calculated.
    :var tuple url_args: URL args to be used while calculating
        the URL using Django's ``reverse``.
    :var dict url_kwargs: URL kwargs to be used while calculating
        the URL using Django's ``reverse``.
    :var str request_method: HTTP method to use for the request.
        Default is ``"post"``
    :var dict data: Data to be sent to the server in the request
    :var platform_utils.utils.dt_context.BaseContext state:
        Reference to a global state from the test runner.
    :var list TestStep.validators: List of response validators
    :var response: Server response for the made request.
        This attribute is only available after
        :py:meth:`request` is called.

    :param kwargs: A dictionary of values which will overwrite
        any instance attributes. This allows to pass additional
        data to the test step without necessarily subclassing
        and manually instantiating step instance.
    :type kwargs: dict
    """
    test = unittest.TestCase('__init__')

    url_name = None
    url_args = None
    url_kwargs = None

    request_method = 'post'

    state = None
    data = None

    validators = []

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def init(self, client, steps, step_index, step_key, state):
        """
        Initialize the step with necessary values from the
        test runner.

        :param client: Django test client to use to make server requests
        :type client: django.test.client.Client

        :param steps: All steps from the test runner.
            This and step index allows to get previous and/or
            next steps.
        :type steps: collections.OrderedDict

        :param step_index: Index of the step within all steps
            test runner will execute.
        :type step_index: int

        :param step_key: Key of the state of how it was provided
            to the test runner in case test step needs
            to reference other steps within the runner by their.
        :type step_key: str

        :param state: Global state reference from the test runner.
        :type state: platform_utils.utils.dt_context.BaseContext
        """
        self.client = client
        self.steps = steps
        self.step_index = step_index
        self.step_key = step_key

        self.state = state

    @property
    def prev_steps(self):
        """
        Get :py:class:`collections.OrderedDict` of previous steps
        excluding itself, if any.

        .. note:: The steps are returned in order of adjacency
            from the current step. For example
            (using list instead of OrederedDict in example)::

                > step = TestStep()
                > step.steps = [0, 1, 2, 3, 4]
                > step.step_index = 3
                > step.prev_steps
                [2, 1, 0]

        :rtype: :py:class:`collections.OrderedDict`
        """
        return OrderedDict(list(self.steps.items())[:self.step_index][::-1])

    @property
    def next_steps(self):
        """
        Get :py:class:`collections.OrderedDict` of next steps
        excluding itself, if any.

        :rtype: :py:class:`collections.OrderedDict`
        """
        return OrderedDict(list(self.steps.items())[self.step_index + 1:])

    @property
    def prev_step(self):
        """
        Get previous step instance, if any.

        :rtype: :py:class:`TestStep`
        """
        prev_steps = list(self.prev_steps.values())
        return prev_steps[0] if prev_steps else None

    @property
    def next_step(self):
        """
        Get previous step instance, if any.

        :rtype: :py:class:`TestStep`
        """
        next_steps = list(self.next_steps.values())
        return next_steps[0] if next_steps else None

    def get_url_args(self):
        """
        Get ``url_args`` which will be used to compute
        the URL using ``reverse``.

        By default this returns :py:attr:`url_args`, if defined,
        else empty tuple.

        :rtype: tuple
        """
        return self.url_args or tuple()

    def get_url_kwargs(self):
        """
        Get ``url_kwargs`` which will be used to compute
        the URL using ``reverse``.

        By default this returns :py:attr:`url_kwargs`, if defined,
        else empty dict.

        :rtype: dict
        """
        return self.url_kwargs or {}

    def get_url(self):
        """
        Compute the URL to request using Django's ``reverse``.

        Reverse is called using :py:attr:`url_name`,
        :py:meth:`get_url_args` and :py:meth:`get_url_args`.

        :rtype: str
        """
        return reverse(self.url_name,
                       args=self.get_url_args(),
                       kwargs=self.get_url_kwargs())

    def get_request_data(self, data=None):
        """
        Get data dict to be sent to the server.

        :param data: Data to be used while sending server request.
            If not defined, :py:attr:`data` is returned.

        :rtype: dict
        """
        return (self.data if data is None else data) or {}

    def get_request_kwargs(self):
        """
        Get kwargs to be passed to the :py:attr:`client`.

        By default this returns dict of format::

            {
                'path': ...,
                'data': ...
            }

        Can be overwritten in case additional parameters need
        to be passed to the client to make the request.

        :rtype: dict
        """
        return {
            'path': self.get_url(),
            'data': self.get_request_data(),
        }

    def get_validators(self):
        """
        Get all validators.

        By default returns :py:attr:`validators` however
        can be used as a hook to returns additional validators
        dynamically.

        :rtype: list
        """
        return self.validators

    def request(self):
        """
        Make the server request. Server response is then saved
        in :py:attr:`request`.

        Before making the request, :py:meth:`pre_request_hook`
        is called and :py:meth:`post_request_hook` is called
        after the request.

        :returns: server response
        """
        self.pre_request_hook()
        try:
            self.response = (getattr(self.client, self.request_method)
                             (**self.get_request_kwargs()))
        except Exception:
            validator = BaseValidator(self)
            e_type, e, e_traceback = sys.exc_info()

            msg = ('{} failed:\n\n{}'
                   ''.format(validator._get_base_error_message(),
                             six.text_type(e)))

            cls = type(e_type.__name__, (Exception,), {})
            six.reraise(
                cls,
                cls(msg),
                e_traceback
            )

        self.post_request_hook()

        return self.response

    def test_response(self):
        """
        Test the server response by looping over all
        validators as returned by :py:meth:`get_validators`.

        Before assertions, :py:meth:`pre_test_response`
        is called and :py:meth:`post_test_response` is called
        after assertions.
        """
        self.pre_test_response()

        for validator in self.get_validators():
            if inspect.isclass(validator):
                validator(self).test(self)
            else:
                validator.test(self)

        self.post_test_response()

    def pre_test_response(self):
        """
        Hook which is executed before validating the response.
        """
        pass

    def post_test_response(self):
        """
        Hook which is executed after validating the response.
        """
        pass

    def pre_request_hook(self):
        """
        Hook which is executed before server request is sent.
        """
        pass

    def post_request_hook(self):
        """
        Hook which is executed after server request is sent.
        """
        pass


class StatefulUrlParamsTestStep(TestStep):

    """
    Test step same as :py:class:`TestStep`
    except it references ``url_args`` and ``url_kwargs`` from the state.

    Having url computed from the state, allows for a particular step
    to change ``url_args`` or ``url_kwargs`` hence future
    steps will fetch different resources.
    """

    def get_url_args(self):
        """
        Get URL args for Django's ``reverse``

        Similar to :py:meth:`TestStep.get_url_args` except
        url args are retrieved by default from state and if not
        available get args from class attribute.

        :returns: tuple of url_args
        """
        args = super(StatefulUrlParamsTestStep, self).get_url_args()
        return self.state.get('url_args', args)

    def get_url_kwargs(self):
        """
        Get URL kwargs fpr Django's ``reverse``

        Similar to :py:meth:`TestStep.get_url_kwargs` except
        url args are retrieved by default from state and if not
        available get kwargs from class attribute.

        :returns: dict of url_kwargs
        """
        kwargs = super(StatefulUrlParamsTestStep, self).get_url_kwargs()
        return self.state.get('url_kwargs', kwargs)



from __future__ import print_function, unicode_literals
from unittest import TestCase

import mock
from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import Resolver404
from django.http.response import HttpResponse
from django.template.response import SimpleTemplateResponse

from subui.validators import (
    BaseValidator,
    FormInitialDataValidator,
    HeaderValidator,
    RedirectToRouteValidator,
    ResponseContentContainsValidator,
    ResponseContentNotContainsValidator,
    SessionDataValidator,
    StatusCodeValidator,
)


TESTING_MODULE = 'subui.validators'


class TestBaseValidator(TestCase):
    def setUp(self):
        class Validator(BaseValidator):
            pass

        self.validator_class = Validator

    def test_init(self):
        """
        Test that __init__ properly stores given parameters
        """
        mock_step = mock.MagicMock(test=mock.create_autospec(self))
        kwargs = {
            'hello': 'world',
            'foo': 'bar',
        }

        self.validator_class.hello = 'mars'
        validator = self.validator_class(mock_step, **kwargs)

        self.assertIs(validator.step, mock_step)
        for key, value in kwargs.items():
            self.assertTrue(hasattr(validator, key))
            self.assertEqual(getattr(validator, key), value)

    def test_get_expected_attrs(self):
        """
        Test that _get_expected_attrs returns expected attributes
        from all base classes.
        """

        class Validator(BaseValidator):
            expected_attrs = ('foo1',)

        class Validator2(Validator):
            expected_attrs = ('foo2',)

        class Validator3(Validator):
            expected_attrs = ('foo3',)

        class Validator4(Validator2, Validator3):
            expected_attrs = ('foo4',)

        validator = Validator4()

        self.assertSetEqual(set(validator._get_expected_attrs()),
                            {'foo1', 'foo2', 'foo3', 'foo4'})

    def test_check_importper_configuration(self):
        """
        Test that _check_improper_configuration verifies
        that all expected attributes are defined.
        """
        self.validator_class.expected_attrs = ('foo',)
        validator = self.validator_class()

        with self.assertRaises(ImproperlyConfigured):
            validator._check_improper_configuration()

        validator.foo = 'bar'
        validator._check_improper_configuration()

        self.assertTrue(True)

    @mock.patch.object(BaseValidator, '_check_improper_configuration')
    def test_test(self, mock_check_improper_configuration):
        """
        Test that test sets the test step and calls the
        _check_improper_configuration
        """
        validator = self.validator_class()
        mock_step = mock.MagicMock(test=mock.create_autospec(self))

        validator.test(mock_step)

        self.assertIs(validator.step, mock_step)
        mock_check_improper_configuration.assert_called_once_with()


class TestHeaderValidator(TestCase):
    def setUp(self):
        self.validator = HeaderValidator()

    def test_get_expected_attrs(self):
        """
        Test that _get_expected_attrs accounts value of
        test_header_value attribute
        """
        self.assertSetEqual(set(self.validator._get_expected_attrs()),
                            {'header_name', 'expected_header'})

        self.validator.test_header_value = False
        self.assertSetEqual(set(self.validator._get_expected_attrs()),
                            {'header_name'})

    def test_test(self):
        """
        Test that expected header is present in the response
        and its value is validated when test_header_value is True
        """
        self.validator.test_header_value = False
        self.validator.header_name = 'header-name'
        self.validator.expected_header = 'some-value'

        self.validator.step = mock_step = mock.Mock(test=mock.create_autospec(self))
        self.validator.step.response = {'header-name': 'header-name'}
        self.validator.step.test = mock.Mock(unsafe=True)

        self.validator.test(mock_step)

        mock_step.test.assertIn.assert_called_once_with(
            'header-name',
            mock_step.response,
            mock.ANY
        )
        self.assertFalse(mock_step.test.assertEqual.called)

        self.validator.test_header_value = True
        mock_step.test.reset_mock()
        self.validator.test(mock_step)

        mock_step.test.assertEqual.assert_called_once_with(
            mock_step.response['header-name'],
            'some-value',
            mock.ANY
        )

        self.validator.test_contains_value = True
        mock_step.test.reset_mock()
        self.validator.test(mock_step)

        self.assertFalse(self.validator.test_header_value)
        mock_step.test.assertIn.assert_has_calls([
            mock.call('header-name',
                      mock_step.response,
                      mock.ANY),
            mock.call('some-value',
                      mock_step.response['header-name'],
                      mock.ANY)
        ])
        self.assertFalse(mock_step.test.assertEqual.called)


class TestStatusCodeValidator(TestCase):
    def setUp(self):
        self.validator = StatusCodeValidator()

    def test_test(self):
        """
        Test that the validator is testing status code of the response
        """
        self.validator.expected_status_code = 200

        mock_step = mock.MagicMock(test=mock.create_autospec(self))
        mock_step.test = mock.Mock(unsafe=True)

        self.validator.test(mock_step)

        mock_step.test.assertEqual.assert_called_once_with(
            mock_step.response.status_code,
            200,
            mock.ANY
        )


class TestRedirectToRouteValidator(TestCase):
    def setUp(self):
        self.validator = RedirectToRouteValidator()

    @mock.patch(TESTING_MODULE + '.resolve')
    def test_test(self, mock_resolve):
        """
        Check that test verified that the response redirects
        to a particular route name
        """
        self.validator.expected_route_name = 'some-route'
        response = HttpResponse('', status=302)
        response['Location'] = 'http://example.com/foo/bar/?query=here#fragment'

        mock_step = mock.MagicMock(
            test=mock.create_autospec(self),
            response=response,
        )
        mock_resolve.return_value.view_name = 'view-name'

        self.validator.test(mock_step)

        mock_step.test.assertEqual.assert_any_call(
            'view-name',
            'some-route',
            mock.ANY
        )
        mock_resolve.assert_called_once_with('/foo/bar/')

    @mock.patch(TESTING_MODULE + '.resolve')
    def test_test_invalid(self, mock_resolve):
        """
        Check that test verified that the response redirects
        to a particular route name when the redirect path
        cannot be resolved by Django
        """
        self.validator.expected_route_name = 'some-route'
        response = HttpResponse('', status=302)
        response['Location'] = 'http://example.com/foo/bar/?query=here#fragment'

        mock_step = mock.MagicMock(
            test=mock.create_autospec(self),
            response=response,
        )
        mock_resolve.side_effect = Resolver404
        mock_step.test.fail.side_effect = RuntimeError

        with self.assertRaises(RuntimeError):
            self.validator.test(mock_step)

        mock_step.test.fail.assert_called_once_with(mock.ANY)
        mock_resolve.assert_called_once_with('/foo/bar/')


class TestResponseContentContainsValidator(TestCase):
    def setUp(self):
        self.validator = ResponseContentContainsValidator()

    def test_test(self):
        self.validator.expected_content = 'some content'

        mock_step = mock.MagicMock(
            response=mock.MagicMock(
                content=b'abc',
            ),
            test=mock.create_autospec(self),
        )

        self.validator.test(mock_step)

        mock_step.test.assertIn.assert_any_call(
            'some content',
            'abc',
            mock.ANY
        )


class TestResponseContentNotContainsValidator(TestCase):
    def setUp(self):
        self.validator = ResponseContentNotContainsValidator()

    def test_test(self):
        self.validator.unexpected_content = 'some content'

        mock_step = mock.MagicMock(
            response=mock.MagicMock(
                content=b'abc',
            ),
            test=mock.create_autospec(self),
        )

        self.validator.test(mock_step)

        mock_step.test.assertNotIn.assert_any_call(
            'some content',
            'abc',
            mock.ANY
        )


class TestSessionDataValidator(TestCase):
    def setUp(self):
        self.validator = SessionDataValidator()

    def test_test(self):
        self.validator.expected_session_key = 'foo'
        self.validator.expected_session_secondary_keys = ['bar']
        session = {
            'foo': {
                'bar': 'value',
            },
        }

        mock_step = mock.Mock(
            response=mock.Mock(
                wsgi_request=mock.Mock(
                    session=session,
                ),
            ),
            test=mock.create_autospec(self),
        )

        self.validator.test(mock_step)

        mock_step.test.assertIn.assert_has_calls([
            mock.call('foo',
                      session.keys(),
                      mock.ANY),
            mock.call('bar',
                      session['foo'].keys(),
                      mock.ANY)
        ])
        mock_step.test.assertIsInstance.assert_called_once_with(
            session['foo'],
            dict,
            mock.ANY
        )
        mock_step.test.assertIsNotNone.assert_called_once_with(
            session['foo']['bar'],
            mock.ANY
        )


class TestFormInitialDataValidator(TestCase):
    def setUp(self):
        self.validator = FormInitialDataValidator()

    def test_test(self):
        self.validator.test_initial_value = True
        self.validator.initial_data_key = 'foo'
        self.validator.expected_initial_data_value = 'bar'

        initial = {
            'foo': 'value',
        }
        mock_step = mock.Mock(
            response=mock.Mock(
                spec=SimpleTemplateResponse,
                context_data={
                    'form': mock.Mock(
                        initial=initial,
                    )
                }
            ),
            test=mock.create_autospec(self),
        )

        self.validator.test(mock_step)

        mock_step.test.assertIsInstance.assert_has_calls([
            mock.call(mock_step.response,
                      SimpleTemplateResponse,
                      mock.ANY),
            mock.call(initial,
                      dict,
                      mock.ANY, )
        ])
        mock_step.test.assertIn.assert_has_calls([
            mock.call('form',
                      mock_step.response.context_data,
                      mock.ANY),
            mock.call('foo',
                      initial.keys(),
                      mock.ANY),
        ])
        mock_step.test.assertIsNotNone.assert_called_once_with(
            'value',
            mock.ANY,
        )
        mock_step.test.assertEqual.assert_called_once_with(
            'value',
            'bar',
            mock.ANY,
        )



"""
SubUI is a framework to ease the pain of writing and running
integration tests. The "SubUI" part means that it is not meant
to test any of the UI (like html validation) but instead
allows to make complete workflow server integration tests.

Introduction
------------

The framework consists of 3 main components:

1.  **SubUI test runner**

    This is the interface layer of the SubUI framework.
    In other words, test methods will instantiate the runner
    and use it to interact with the SubUI integration tests.
    Primary job of a SubUI test runner is to execute test
    steps (described below) in correct order and maintain
    state between steps if necessary.

    .. note:: Even though it is called test runner, it does not
       replace or even relate to ``nosetests`` test runner.

2.  **Test Steps**

    In the most part, this framework is meant to make integration
    tests for complete workflows (e.g. go to page 1 ->
    submit form and assert redirect to page 2 -> go to page 2).
    A test step is a self-contained piece of the complete workflow
    like "go to page 1". Combinations of multiple steps then make up
    the workflow. Since steps are independent, they should know how to
    complete their task (e.g. submit form via ``POST``) and validate
    that they got expected result from the server. To allow
    flexible validation, they themselves do not validate anything
    but use validators (described below) to inspect server response
    in very similar way to how Django Form Field uses validators to
    verify user-input.

3.  **Validator**

    Validator's task is to make assertions about the response from
    the server. All validators are pretty straight forwards like
    assert that the response status code is ``Redirect - 302``
    or that redirect header ``Location`` is returned. More complex
    assertions can be made by either using multiple validators
    in each test step or make more complex validator via
    multiple class inheritance.

Example
-------

An example should show some of the advantages of using this
framework for a hypothetical todo application:

.. code-block:: python
    :linenos:

    # define steps
    class GoToLogin(TestStep):
        url_name = 'login'
        request_method = 'get'
        validators = [StatusCodeOkValidator]

    class SubmitLoginForm(TestStep):
        url_name = 'form'
        validators = [RedirectToRouteValidator(expected_route_name='list')]
        data = {
            'username': 'user',
            'password': 'password',
        }

    class GoToList(TestStep):
        url_name = 'list'
        request_method = 'get'
        validators = [StatusCodeOkValidator]

    class CreateToDo(TestStep):
        url_name = 'create'
        validators = [RedirectToRouteValidator(expected_route_name='list')]
        data = {
            'notes': 'need to finish something',
            'due date': '2015-01-01',
        }

    # integration tests
    class TestWorkflow(TestCase):
        def test_login_and_create_todo(self):
            runner = SubUITestRunner(
                OrderedDict((
                    ('login', GoToLogin),
                    ('login_submit', SubmitLoginForm),
                    ('list1', GoToList),
                    ('create', CreateToDo),
                    ('list2', GoToList),
                )),
                client=self.client
            ).run()

            self.assertNotContains(runner.steps['list1'].response,
                                   'need to finish something')
            self.assertContains(runner.steps['list2'].response,
                                'need to finish something')

        def test_just_create(self):
            data = {
                'notes': 'other task here to complete',
                'due date': '2015-01-01',
            }
            runner = SubUITestRunner(
                OrderedDict((
                    ('list1', GoToList),
                    ('create', CreateToDo(data=data)),
                    ('list2', GoToList),
                )),
                client=self.client
            ).run()

            self.assertNotContains(runner.steps['list1'].response,
                                   'other task here to complete')
            self.assertContains(runner.steps['list2'].response,
                                'other task here to complete')

some useful things to note about what happened above:

* Reuse of test steps. Since each step is self-contained,
  they can be combined in different ways to make
  different integration tests. They can even be reused
  multiple times within the same integration test.
* Step attributes can easily be overwritten if need to like
  in ``test_just_create`` test method - ``CreateToDo``'s
  data is overwritten to post different values.
* Assertions on steps can be performed outside of the test
  runner. After steps are executed, all steps can be accessed
  via ``runner.steps`` attribute which will be an instance
  of :py:class:`collections.OrderedDict`.
* :py:class:`subui.validators.RedirectToRouteValidator`
  is combined validator via multiple inheritance which
  verifies that the response status code is 302 - Redirect;
  ``Location`` response header is present; and that
  the page redirects to a particular route as determined
  by Django's ``resolve``.

Advanced Use-Cases
------------------

More advanced things can be accomplished with the framework.
In the previous example, all steps had a fixed url without any
parameters. This example will use state to pass information
between steps:

.. code-block:: python
    :linenos:

    # login returns redirect to user profile with user id
    class LoginStep(TestStep):
        url_name = 'login'
        validators = [RedirectToRouteValidator(expected_route_name='profile')]
        data = {
            'username': 'username',
            'password': 'password',
        }

        def post_test_response(self):
            # extract user kwargs from redirect location
            resolved = resolve(self.response['Location'])
            self.state.push({
                'url_kwargs': resolved.kwargs,
            })

    class ProfileStep(StatefulUrlParamsTestStep):
        url_name = 'profile'  # requires url kwargs of username
        request_method = 'get'
        validators = [StatusCodeOkValidator]

    class EditProfileStep(StatefulUrlParamsTestStep):
        url_name = 'edit_profile'  # requires url kwarg of username
        validators = [StatusCodeOkValidator]

    class TestWorkflow(TestCase):
        def test_login_and_edit(self):
            data = {
                'username': 'otheruser',
                'password': 'otherpassword',
            }
            runner = SubUITestRunner(
                [
                    LoginStep,                   # 0
                    ProfileStep,                 # 1
                    EditProfileStep(data=data),  # 2
                    ProfileStep,                 # 3
                ],
                client=self.client
            ).run()

            self.assertContains(runner.steps[3].response,
                                'otheruser')

some notes about what happened:

* ``LoginStep`` uses a hook
  :py:meth:`subui.step.TestStep.post_test_response`
  to add data to a state. Since state is global for all steps within
  test runner, other steps can access it.
* ``ProfileStep`` and ``EditProfileStep`` subclass
  :py:class:`subui.step.StatefulUrlParamsTestStep`
  which uses state to get url args and kwargs.
* When using ``resolve`` in ``post_test_response``, there is no
  need to do ``try: except Resolver404`` since that will be executed
  after validator verifications hence it is guaranteed that
  the url will resolve without issues.
* Steps are provided as :py:func:`list` instead of
  :py:class:`collections.OrderedDict`. Test runner automatically converts
  the steps into :py:class:`collections.OrderedDict` with keys as indexes
  which allows to type test runner a bit faster ;-) in case you don't
  need to reference steps with particular keys.
"""
from __future__ import unicode_literals


__author__ = 'Miroslav Shubernetskiy'
__version__ = '0.1.0'
__description__ = 'Python dict with stacked context data'



from __future__ import print_function, unicode_literals
from collections import OrderedDict
from unittest import TestCase

import mock
import six

from subui.step import StatefulUrlParamsTestStep, TestStep

from .utils import patch_class_method_with_original


TESTING_MODULE = 'subui.step'


class TestTestStep(TestCase):
    def test__init__(self):
        """
        Test that __init__ overwrites attributes with
        given parameters
        """
        step = TestStep(hello='world')
        self.assertTrue(hasattr(step, 'hello'))
        self.assertEqual(getattr(step, 'hello'), 'world')

    def test_init(self):
        """
        Test that init stores given parameters as attributes
        """
        step = TestStep()
        step.init(mock.sentinel.client,
                  mock.sentinel.steps,
                  mock.sentinel.step_index,
                  mock.sentinel.step_key,
                  mock.sentinel.state)

        self.assertEqual(step.client, mock.sentinel.client)
        self.assertEqual(step.steps, mock.sentinel.steps)
        self.assertEqual(step.step_index, mock.sentinel.step_index)
        self.assertEqual(step.step_key, mock.sentinel.step_key)
        self.assertEqual(step.state, mock.sentinel.state)

    def test_steps(self):
        """
        Test prev_steps, next_steps, prev_step and next_step
        properties
        """
        step = TestStep()
        step.steps = OrderedDict((
            ('one', 1),
            ('two', 2),
            ('three', 3),
            ('four', 4),
            ('five', 5),
        ))
        step.step_index = 2

        self.assertEqual(
            step.prev_steps,
            OrderedDict((
                ('two', 2),
                ('one', 1),
            ))
        )
        self.assertEqual(
            step.prev_step,
            2,
        )
        self.assertEqual(
            step.next_step,
            4,
        )

    def test_get_url_args(self):
        """
        Test that get_url_args returns url args if defined
        """
        step = TestStep()
        self.assertTupleEqual(step.get_url_args(), tuple())

        step.url_args = ('foo',)
        self.assertTupleEqual(step.get_url_args(), ('foo',))

    def test_get_url_kwargs(self):
        """
        Test that get_url_kwargs returns url kwargs if defined
        """
        step = TestStep()
        self.assertDictEqual(step.get_url_kwargs(), {})

        step.url_kwargs = {'foo': 'bar'}
        self.assertDictEqual(step.get_url_kwargs(), {'foo': 'bar'})

    @patch_class_method_with_original(TestStep, 'get_url_kwargs')
    @patch_class_method_with_original(TestStep, 'get_url_args')
    @mock.patch(TESTING_MODULE + '.reverse')
    def test_get_url(self,
                     mock_get_url_args,
                     mock_get_url_kwargs,
                     mock_reverse):
        mock_reverse.return_value = mock.sentinel.reverse

        step = TestStep()
        step.url_name = 'url-name'
        actual = step.get_url()

        self.assertEqual(actual, mock.sentinel.reverse)
        mock_get_url_args.assert_called_once_with(step)
        mock_get_url_kwargs.assert_called_once_with(step)
        mock_reverse.assert_called_once_with(
            'url-name', args=tuple(), kwargs={})

    def test_get_request_data(self):
        """
        Test that get_request_data returns data if defined
        """
        step = TestStep()
        actual = step.get_request_data()
        self.assertEqual(actual, {})

        step.data = {'hello': 'world'}
        actual = step.get_request_data()
        self.assertEqual(actual, {'hello': 'world'})

        actual = step.get_request_data({'foo': 'bar'})
        self.assertEqual(actual, {'foo': 'bar'})

    def test_get_validators(self):
        """
        Test that get_validators returns validators
        """
        step = TestStep()
        step.validators = mock.sentinel.validators
        self.assertEqual(step.get_validators(), mock.sentinel.validators)

    @patch_class_method_with_original(TestStep, 'post_request_hook')
    @patch_class_method_with_original(TestStep, 'pre_request_hook')
    @mock.patch.object(TestStep, 'get_url')
    def test_request(self,
                     mock_pre_request_hook,
                     mock_post_response_hook,
                     mock_get_url):
        """
        Test that request correctly calls the client
        """
        mock_client = mock.MagicMock()
        mock_client.methodname.return_value = mock.sentinel.response
        mock_get_url.return_value = mock.sentinel.url

        step = TestStep(client=mock_client, data={'hello': 'world'})
        step.request_method = 'methodname'

        actual = step.request()

        self.assertEqual(actual, mock.sentinel.response)
        mock_pre_request_hook.assert_called_once_with(step)
        mock_post_response_hook.assert_called_once_with(step)

    @mock.patch.object(TestStep, 'get_url')
    def test_request_with_error(self, mock_get_url):
        """
        Test that request exceptions are handled correctly
        """
        mock_client = mock.MagicMock()
        mock_client.methodname.side_effect = ValueError
        mock_get_url.return_value = mock.sentinel.url

        step = TestStep(client=mock_client, step_key='foo', data={'hello': 'world'})
        step.request_method = 'methodname'

        with self.assertRaises(Exception) as e:
            step.request()

        self.assertEqual(e.exception.__class__.__name__, 'ValueError')
        self.assertEqual(
            six.text_type(e.exception).splitlines()[0],
            'Response for {foo:TestStep} requesting "sentinel.url" failed:'
        )

    @patch_class_method_with_original(TestStep, 'post_test_response')
    @patch_class_method_with_original(TestStep, 'pre_test_response')
    @mock.patch('inspect.isclass')
    def test_test_response(self,
                           mock_pre_test_response,
                           mock_post_test_response,
                           mock_isclass):
        """
        Test that test_response loops over all validators
        and correctly calls them
        """
        mock_isclass.side_effect = [True, False]
        validator = mock.MagicMock()
        class_validator = mock.MagicMock()
        step = TestStep(validators=[class_validator, validator])
        step.test_response()

        validator.test.assert_called_once_with(step)
        class_validator.assert_called_once_with(step)
        class_validator().test.assert_called_once_with(step)
        mock_pre_test_response.assert_called_once_with(step)
        mock_post_test_response.assert_called_once_with(step)


class TestStatefulUrlParamsTestStep(TestCase):
    @mock.patch.object(TestStep, 'get_url_args')
    def test_get_url_args(self, mock_get_url_args):
        """
        Test that get_url_args returns args from the state if present
        """
        mock_get_url_args.return_value = mock.sentinel.super
        mock_state = mock.MagicMock()
        mock_state.get.return_value = mock.sentinel.args

        step = StatefulUrlParamsTestStep(state=mock_state)
        actual = step.get_url_args()

        self.assertEqual(actual, mock.sentinel.args)
        mock_state.get.assert_called_once_with('url_args', mock.sentinel.super)

    @mock.patch.object(TestStep, 'get_url_kwargs')
    def test_get_url_kwargs(self, mock_get_url_kwargs):
        """
        Test that get_url_args returns kwargs from the state if present
        """
        mock_get_url_kwargs.return_value = mock.sentinel.super
        mock_state = mock.MagicMock()
        mock_state.get.return_value = mock.sentinel.kwargs

        step = StatefulUrlParamsTestStep(state=mock_state)
        actual = step.get_url_kwargs()

        self.assertEqual(actual, mock.sentinel.kwargs)
        mock_state.get.assert_called_once_with('url_kwargs', mock.sentinel.super)
